from typing import Literal
from .base import Color
from tkinter import Tk

def rgb_to_hex(r: int, g: int, b: int) -> str:
    return f"#{r:02x}{g:02x}{b:02x}"

def color_text(
    text: str,
    color: Color,
    *,
    mode: Literal["text", "print"] = "text",
    allowed_formats: list[str] = ["rgb", "rgba", "hsl"],
    **kw
):
    # Validate allowed formats
    for fmt in allowed_formats:
        if fmt not in ["rgb", "rgba", "hsl"]:
            raise ValueError(f"wrong format '{fmt}'")
    if color.format not in allowed_formats:
        raise ValueError(f"wrong format '{color.format}'")

    rgb_color = color.to_rgb()
    r, g, b = rgb_color.r, rgb_color.g, rgb_color.b
    gentext = f"\033[38;2;{r};{g};{b}m{text}\033[0m"

    if mode == "text":
        return gentext
    elif mode == "print":
        print(gentext)

def display(
    color: Color,
    *,
    allowed_formats: list[str] = ["rgb", "rgba", "hsl"],
    **kw
):
    # Validate allowed formats
    for fmt in allowed_formats:
        if fmt not in ["rgb", "rgba", "hsl"]:
            raise ValueError(f"wrong format '{fmt}'")
    if color.format not in allowed_formats:
        raise ValueError(f"wrong format '{color.format}'")

    if color.format == "rgb":
        root = Tk()
        root.configure(bg=rgb_to_hex(color.r, color.g, color.b))
        root.mainloop()
    elif color.format == "rgba":
        root = Tk()
        root.configure(bg=rgb_to_hex(color.r, color.g, color.b))
        root.attributes("-alpha", color.alpha)
        root.mainloop()
    elif color.format == "hsl":
        display(color.to_rgb())


from typing import Literal as _Literal
from math import floor as _floor

class ColorError(Exception):
    pass

class Color:
    def __init__(self, color: tuple, color_format: _Literal["rgb", "rgba", "hsl"] = "rgb"):
        if len(color) != len(color_format):
            raise ColorError("tuple not matching with format")
        wrong_types = {type(obj).__name__ for obj in color if not isinstance(obj, (int, float))}
        if wrong_types:
            raise IndexError(f"wrong values in color: {', '.join(wrong_types)}")
        if color_format == "hsl":
            h, s, l = color
            if not (0 <= h <= 360 and 0 <= s <= 1 and 0 <= l <= 1):
                raise ColorError("color not matching with hsl")
        elif color_format in ["rgb", "rgba"]:
            for i, num in enumerate(color):
                if i == 3:
                    if not (0 <= num <= 1):
                        raise ColorError(f"tuple not matching with {color_format}")
                else:
                    if not (0 <= num <= 255):
                        raise ColorError(f"tuple not matching with {color_format}")
        if color_format in ("rgb", "rgba"):
            self.r, self.g, self.b = color[0], color[1], color[2]
            if color_format == "rgba":
                self.alpha = color[3]
        else:
            self.h, self.s, self.l = color
        self.format = color_format

    def __str__(self):
        if self.format == "rgb":
            return f"rgb({self.r}, {self.g}, {self.b})"
        elif self.format == "rgba":
            return f"rgba({self.r}, {self.g}, {self.b}, {self.alpha})"
        elif self.format == "hsl":
            return f"hsl({self.h}, {self.s}, {self.l})"

    def __repr__(self):
        return f"<Color {self}>"

    def _hsl_to_rgb_values(self):
        c = (1 - abs(2 * self.l - 1)) * self.s
        h_prime = self.h / 60
        x = c * (1 - abs(h_prime % 2 - 1))
        m = self.l - c / 2
        if 0 <= self.h < 60:
            r1, g1, b1 = c, x, 0
        elif 60 <= self.h < 120:
            r1, g1, b1 = x, c, 0
        elif 120 <= self.h < 180:
            r1, g1, b1 = 0, c, x
        elif 180 <= self.h < 240:
            r1, g1, b1 = 0, x, c
        elif 240 <= self.h < 300:
            r1, g1, b1 = x, 0, c
        else:
            r1, g1, b1 = c, 0, x
        r = _floor((r1 + m) * 255)
        g = _floor((g1 + m) * 255)
        b = _floor((b1 + m) * 255)
        return r, g, b

    def to_rgb(self, *, inplace: bool = False, **kw):
        if self.format == "rgb":
            return self if not inplace else None
        elif self.format == "rgba":
            if inplace:
                del self.alpha
                self.format = "rgb"
            else:
                return Color((self.r, self.g, self.b), "rgb")
        elif self.format == "hsl":
            r, g, b = self._hsl_to_rgb_values()
            if inplace:
                self.r, self.g, self.b = r, g, b
                self.format = "rgb"
                del self.h, self.s, self.l
            else:
                return Color((r, g, b), "rgb")

    def to_rgba(self, *, inplace: bool = False, **kw):
        if self.format == "rgba":
            return self if not inplace else None
        elif self.format == "rgb":
            if inplace:
                self.alpha = 1.0
                self.format = "rgba"
            else:
                return Color((self.r, self.g, self.b, 1.0), "rgba")
        elif self.format == "hsl":
            r, g, b = self._hsl_to_rgb_values()
            if inplace:
                self.r, self.g, self.b, self.alpha = r, g, b, 1.0
                self.format = "rgba"
                del self.h, self.s, self.l
            else:
                return Color((r, g, b, 1.0), "rgba")

    def to_hsl(self, *, inplace: bool = False, **kw):
        if self.format == "hsl":
            return self if not inplace else None
        r1 = self.r / 255
        g1 = self.g / 255
        b1 = self.b / 255
        cmax = max(r1, g1, b1)
        cmin = min(r1, g1, b1)
        delta = cmax - cmin

        if delta == 0:
            h = 0
        elif cmax == r1:
            h = (g1 - b1) / delta % 6
        elif cmax == g1:
            h = (b1 - r1) / delta + 2
        else:
            h = (r1 - g1) / delta + 4
        h *= 60

        l = (cmax + cmin) / 2
        s = 0 if delta == 0 else delta / (1 - abs(2 * l - 1))

        if inplace:
            self.h, self.s, self.l = h, s, l
            del self.r, self.g, self.b
            if self.format == "rgba":
                del self.alpha
            self.format = "hsl"
        else:
            return Color((h, s, l), "hsl")



from .base import Color, ColorError
from math import sqrt

def average(num1: int, num2: int):
    return (num1 + num2) / 2

def mix(color1: Color, color2: Color):
    if color1.format != color2.format:
        color1 = color1.to_rgb()
        color2 = color2.to_rgb()
    if color1.format == "rgb":
        return Color((average(color1.r, color2.r), average(color1.g, color2.g), average(color1.b, color2.b)))
    elif color1.format == "rgba":
        return Color(
            (average(color1.r, color2.r), average(color1.g, color2.g), average(color1.b, color2.b), max(color1.alpha, color2.alpha)),
            "rgba"
        )
    else:
        return mix(color1.to_rgb(), color2.to_rgb())

def invert(color: Color):
    if color.format == "rgb":
        return Color((255 - color.r, 255 - color.g, 255 - color.b))
    elif color.format == "rgba":
        return Color((255 - color.r, 255 - color.g, 255 - color.b, color.alpha), "rgba")
    else:
        return invert(color.to_rgb())

def put_ratio(color: Color, ratio: float):
    if color.format == "rgb":
        color.r = min(255, color.r * ratio)
        color.g = min(255, color.g * ratio)
        color.b = min(255, color.b * ratio)
        return color
    elif color.format == "rgba":
        new_rgb = put_ratio(color.to_rgb(), ratio)
        new_rgb.to_rgba(inplace=True)
        new_rgb.alpha = color.alpha
        return new_rgb
    else:
        return put_ratio(color.to_rgb(), ratio)

def adjust(c):
    c = c / 255  # normalize
    return c / 12.92 if c <= 0.03928 else ((c + 0.055) / 1.055) ** 2.4

def is_dark(color: Color):
    color = color.to_rgb() if color.format != "rgb" else color
    newr, newg, newb = adjust(color.r), adjust(color.g), adjust(color.b)
    return (0.2126 * newr + 0.7152 * newg + 0.0722 * newb) < 0.5

def is_light(color: Color):
    return not is_dark(color)

def distance(color1: Color, color2: Color):
    color1 = color1.to_rgb() if color1.format != "rgb" else color1
    color2 = color2.to_rgb() if color2.format != "rgb" else color2
    return sqrt((color1.r - color2.r) ** 2 + (color1.g - color2.g) ** 2 + (color1.b - color2.b) ** 2)


from .base import ColorError, Color
from .presets import PRESET_RED_RGB, PRESET_ORANGE_RGB, PRESET_YELLOW_RGB, PRESET_GREEN_RGB, PRESET_BLUE_RGB, PRESET_PURPLE_RGB, PRESET_VIOLET_RGB, PRESET_PINK_RGB, PRESET_BROWN_RGB, PRESET_GRAY_RGB, PRESET_WHITE_RGB, PRESET_BLACK_RGB, PRESET_RED_RGBA, PRESET_ORANGE_RGBA, PRESET_YELLOW_RGBA, PRESET_GREEN_RGBA, PRESET_BLUE_RGBA, PRESET_PURPLE_RGBA, PRESET_VIOLET_RGBA, PRESET_PINK_RGBA, PRESET_BROWN_RGBA, PRESET_GRAY_RGBA, PRESET_WHITE_RGBA, PRESET_BLACK_RGBA, PRESET_RED_HSL, PRESET_ORANGE_HSL, PRESET_YELLOW_HSL, PRESET_GREEN_HSL, PRESET_BLUE_HSL, PRESET_PURPLE_HSL, PRESET_VIOLET_HSL, PRESET_PINK_HSL, PRESET_BROWN_HSL, PRESET_GRAY_HSL, PRESET_WHITE_HSL, PRESET_BLACK_HSL
from .printkit import color_text, display
from .modify import mix, invert, put_ratio, is_dark, is_light, distance

__all__ = [
    "ColorError", "Color",
    "PRESET_RED_RGB", "PRESET_ORANGE_RGB", "PRESET_YELLOW_RGB", "PRESET_GREEN_RGB",
    "PRESET_BLUE_RGB", "PRESET_PURPLE_RGB", "PRESET_VIOLET_RGB", "PRESET_PINK_RGB",
    "PRESET_BROWN_RGB", "PRESET_GRAY_RGB", "PRESET_WHITE_RGB", "PRESET_BLACK_RGB",
    "PRESET_RED_RGBA", "PRESET_ORANGE_RGBA", "PRESET_YELLOW_RGBA", "PRESET_GREEN_RGBA",
    "PRESET_BLUE_RGBA", "PRESET_PURPLE_RGBA", "PRESET_VIOLET_RGBA", "PRESET_PINK_RGBA",
    "PRESET_BROWN_RGBA", "PRESET_GRAY_RGBA", "PRESET_WHITE_RGBA", "PRESET_BLACK_RGBA",
    "PRESET_RED_HSL", "PRESET_ORANGE_HSL", "PRESET_YELLOW_HSL", "PRESET_GREEN_HSL",
    "PRESET_BLUE_HSL", "PRESET_PURPLE_HSL", "PRESET_VIOLET_HSL", "PRESET_PINK_HSL",
    "PRESET_BROWN_HSL", "PRESET_GRAY_HSL", "PRESET_WHITE_HSL", "PRESET_BLACK_HSL",
    "color_text", "display",
    "mix", "invert", "put_ratio", "is_dark", "is_light", "distance"
]


from .base import Color

PRESET_RED_RGB = Color((255, 0, 0), "rgb")
PRESET_ORANGE_RGB = Color((255, 165, 0), "rgb")
PRESET_YELLOW_RGB = Color((255, 255, 0), "rgb")
PRESET_GREEN_RGB = Color((0, 128, 0), "rgb")
PRESET_BLUE_RGB = Color((0, 0, 255), "rgb")
PRESET_PURPLE_RGB = Color((128, 0, 128), "rgb")
PRESET_VIOLET_RGB = Color((143, 0, 255), "rgb")
PRESET_PINK_RGB = Color((255, 192, 203), "rgb")
PRESET_BROWN_RGB = Color((165, 42, 42), "rgb")
PRESET_GRAY_RGB = Color((128, 128, 128), "rgb")
PRESET_WHITE_RGB = Color((255, 255, 255), "rgb")
PRESET_BLACK_RGB = Color((0, 0, 0), "rgb")
PRESET_RED_RGBA = Color((255, 0, 0, 1.0), "rgba")
PRESET_ORANGE_RGBA = Color((255, 165, 0, 1.0), "rgba")
PRESET_YELLOW_RGBA = Color((255, 255, 0, 1.0), "rgba")
PRESET_GREEN_RGBA = Color((0, 128, 0, 1.0), "rgba")
PRESET_BLUE_RGBA = Color((0, 0, 255, 1.0), "rgba")
PRESET_PURPLE_RGBA = Color((128, 0, 128, 1.0), "rgba")
PRESET_VIOLET_RGBA = Color((143, 0, 255, 1.0), "rgba")
PRESET_PINK_RGBA = Color((255, 192, 203, 1.0), "rgba")
PRESET_BROWN_RGBA = Color((165, 42, 42, 1.0), "rgba")
PRESET_GRAY_RGBA = Color((128, 128, 128, 1.0), "rgba")
PRESET_WHITE_RGBA = Color((255, 255, 255, 1.0), "rgba")
PRESET_BLACK_RGBA = Color((0, 0, 0, 1.0), "rgba")
PRESET_RED_HSL = Color((0, 1, 0.5), "hsl")
PRESET_ORANGE_HSL = Color((30, 1, 0.5), "hsl")
PRESET_YELLOW_HSL = Color((60, 1, 0.5), "hsl")
PRESET_GREEN_HSL = Color((120, 1, 0.25), "hsl")
PRESET_BLUE_HSL = Color((240, 1, 0.5), "hsl")
PRESET_PURPLE_HSL = Color((300, 1, 0.25), "hsl")
PRESET_VIOLET_HSL = Color((270, 1, 0.5), "hsl")
PRESET_PINK_HSL = Color((350, 1, 0.88), "hsl")
PRESET_BROWN_HSL = Color((30, 0.59, 0.41), "hsl")
PRESET_GRAY_HSL = Color((0, 0, 0.5), "hsl")
PRESET_WHITE_HSL = Color((0, 0, 1), "hsl")
PRESET_BLACK_HSL = Color((0, 0, 0), "hsl")